# ----------------------------------------------------------------------------
# PyGMTSAR
# 
# This file is part of the PyGMTSAR project: https://github.com/mobigroup/gmtsar
# 
# Copyright (c) 2024, Alexey Pechnikov
# 
# Licensed under the BSD 3-Clause License (see LICENSE for details)
# ----------------------------------------------------------------------------
from .datagrid import datagrid
from .tqdm_joblib import tqdm_joblib

class AWS(datagrid, tqdm_joblib):
    # https://copernicus-dem-90m.s3.eu-central-1.amazonaws.com
    base_url = 'https://copernicus-dem-{resolution}m.s3.amazonaws.com'
    # Copernicus_DSM_COG_10_S16_00_W043_00_DEM
    tile_id = 'Copernicus_DSM_COG_{product1}0_{SN2}_00_{WE3}_00_DEM'

    def download_dem(self, geometry, filename=None, n_jobs=-1, product='1s', skip_exist=True):
        """
        Download Copernicus GLO-30/GLO-90 Digital Elevation Model from open AWS storage.

        from pygmtsar import AWS
        dem = AWS().download_dem(AOI.geometry[0])
        dem.plot.imshow()
        """
        import xarray as xr
        import rioxarray
        import geopandas as gpd
        import numpy as np
        from tqdm.auto import tqdm
        import joblib
        import requests
        import io
        import os

        assert product in ['1s', '3s'], f'ERROR: product name is invalid: {product}. Expected names are "1s", "3s"'

        if filename is not None and os.path.exists(filename) and skip_exist:
            print ('NOTE: DEM file exists, ignore the command. Use "skip_exist=False" or omit the filename to allow new downloading')
            return

        lon_start, lat_start, lon_end, lat_end = self.get_bounds(geometry)

        def job_tile(product, lon, lat):
            base_url = self.base_url.format(resolution=30 if product=='1s' else 90)
            tile_id = self.tile_id.format(product1=int(product[0]),
                                          SN2=f'{"S" if lat<0 else "N"}{abs(lat):02}',
                                          WE3=f'{"W" if lon<0 else "E"}{abs(lon):03}')
            url = f'{base_url}/{tile_id}/{tile_id}.tif'
            response = requests.get(url)
            #response.raise_for_status()
            # offshore tiles are missed by design
            if response.status_code != 200:
                return None
            with io.BytesIO(response.content) as f:
                tile = rioxarray.open_rasterio(f, cache=True)\
                    .squeeze(drop=True)\
                    .rename({'y': 'lat', 'x': 'lon'})\
                    .drop_vars('spatial_ref')\
                    .load()
                if tile.lat.diff('lat')[0].item() < 0:
                    tile = tile.reindex(lat=tile.lat[::-1])
            return tile

        left, right = int(np.floor(lon_start)), int(np.floor(lon_end))
        lower, upper = int(np.floor(lat_start)), int(np.floor(lat_end))
        #print ('left, right', left, right, 'lower, upper', lower, upper)
        with self.tqdm_joblib(tqdm(desc='DEM Tile Downloading', total=(right-left+1)*(upper-lower+1))) as progress_bar:
            tile_xarrays = joblib.Parallel(n_jobs=n_jobs)(joblib.delayed(job_tile)(product, x, y)\
                                for x in range(left, right + 1) for y in range(lower, upper + 1))
        dem = xr.combine_by_coords([tile for tile in tile_xarrays if tile is not None])

        if filename is not None:
            if os.path.exists(filename):
                os.remove(filename)
            encoding = {'dem': self._compression(dem.shape)}
            dem.rename('dem').to_netcdf(filename, encoding=encoding, engine=self.netcdf_engine)
        else:
            return dem



# ----------------------------------------------------------------------------
# PyGMTSAR
# 
# This file is part of the PyGMTSAR project: https://github.com/mobigroup/gmtsar
# 
# Copyright (c) 2024, Alexey Pechnikov
# 
# Licensed under the BSD 3-Clause License (see LICENSE for details)
# ----------------------------------------------------------------------------
from .tqdm_joblib import tqdm_joblib
from .S1 import S1

class ESA(tqdm_joblib):
    from datetime import timedelta, datetime

    # Default URL endpoint for performing user authentication with CDSE
    token_url = 'https://identity.dataspace.copernicus.eu/auth/realms/CDSE/protocol/openid-connect/token'
    template_query = (
        "startswith(Name,'{mission_id}') and contains(Name,'{orbit_type}') "
        "and ContentDate/Start lt '{start_time}' and ContentDate/End gt '{stop_time}'"
    )
    # Default URL endpoint for the Copernicus Data Space Ecosystem (CDSE) query REST service
    query_url = 'https://catalogue.dataspace.copernicus.eu/odata/v1/Products'
    # Default URL endpoint for CDSE download REST service
    download_url = 'https://zipper.dataspace.copernicus.eu/odata/v1/Products'
    #pattern_orbit  = r'S1\w_OPER_AUX_\w{3}ORB_OPOD_\d{8}T\d{6}_V\d{8}T\d{6}_\d{8}T\d{6}.EOF'
    #pattern_orbit  = r'S1\w_OPER_AUX_(POE|RES)ORB_OPOD_\d{8}T\d{6}_V\d{8}T\d{6}_\d{8}T\d{6}.EOF'
    #template_orbit = '{mission}_OPER_AUX_(POE|RES)ORB_OPOD_\d+T\d+_V{date_start}T\d+_{date_end}T\d+.EOF'
    # see _select_orbit.py in sentineleof package
    #Orbital period of Sentinel-1 in seconds
    #T_ORBIT = (12 * 86400.0) / 175.0
    resorb_start_offset = timedelta(seconds=(12 * 86400.0) // 175.0 + 60)
    resorb_end_offset = timedelta(seconds=300)
    poeorb_start_offset = timedelta(seconds=(12 * 86400.0) // 175.0 + 60)
    poeorb_end_offset = timedelta(seconds=300)
    
    def __init__(self, username: str = None, password: str= None):
        """
        Register your ESA Copernicus datastore account at https://dataspace.copernicus.eu/
        """
        import asf_search
        import getpass
        if username is None:
            username = getpass.getpass('Please enter your Copernicus Data Space username and press Enter key:')
        if password is None:
            password = getpass.getpass('Please enter your Copernicus Data Space password and press Enter key:')
        self.username = username
        self.password = password

    def _get_copernicus_access_token(self) -> str:
        """
        Authenticate with the Copernicus Data Space Identity Server to obtain an access token.

        Parameters:
        - username: Copernicus Data Space username.
        - password: Copernicus Data Space password.

        Returns:
        - A string containing the access token.
        
        Example:
        username = 'your_copernicus_username'
        password = 'your_copernicus_password'
        access_token = get_copernicus_access_token(username, password)
        print("Access Token:", access_token)
        """
        import requests
        
        # Endpoint for obtaining the access token
        data = {
            'client_id': 'cdse-public',
            'username': self.username,
            'password': self.password,
            'grant_type': 'password',
        }
        try:
            response = requests.post(self.token_url, data=data)
            response.raise_for_status()
            access_token = response.json().get('access_token')
            if not access_token:
                raise Exception('Failed to obtain access token.')
            return access_token
        except requests.exceptions.RequestException as e:
            raise Exception(f'Authentication failed: {e}')

    def _download_orbit(self, access_token : str, time: datetime,
                       mission_id: str, basedir: str) -> bool:
        """
        Queries and downloads RESORB/POEORB files for the given satellite within the specified time range.
        """
        import requests
        import os
        
        for product_type in ['AUX_POEORB', 'AUX_RESORB']:
            if product_type == 'AUX_RESORB':
                orbit_start_offset = self.resorb_start_offset
                orbit_end_offset = self.resorb_end_offset
            elif product_type == 'AUX_POEORB':
                orbit_start_offset = self.poeorb_start_offset
                orbit_end_offset = self.poeorb_end_offset
            else:
                raise Exception(f'Invalid orbit type: {product_type}.')

            start_time = time - orbit_start_offset
            end_time = time + orbit_end_offset
            
            query = self.template_query.format(
                start_time = start_time.strftime('%Y-%m-%dT%H:%M:%SZ'),
                stop_time  = end_time.strftime('%Y-%m-%dT%H:%M:%SZ'),
                mission_id = mission_id,
                orbit_type = product_type,
            )
            #print ('query', query)
        
            # Query the Copernicus Data Space for available orbit files
            headers = {'Authorization': f'Bearer {access_token}'}
            response = requests.get(self.query_url, params={'$filter': query}, headers=headers)
            response.raise_for_status()

            # Process the query results
            results = response.json().get('value', [])
            if not results:
                print(f'No {product_type} files found to cover interval {start_time} - {end_time}')
                continue
        
            if len(results) > 1:
                print(f'Found multiple {product_type} orbits for interval {start_time} - {end_time}', results)
            # TODO: what is the sorting for multiple orbits?
            result = results[0]
            
            # Download the orbit file
            orbit_file_id = result['Id']
            orbit_file_name = result['Name']
            download_url = f'{self.download_url}({orbit_file_id})/$value'
            response = requests.get(download_url, headers=headers, stream=True)
            response.raise_for_status()
            file_path = os.path.join(basedir, orbit_file_name)
            with open(file_path, 'wb') as f:
                for chunk in response.iter_content(chunk_size=256*1024):
                    f.write(chunk)            
            return orbit_file_id

        return

    def download_orbits(self, basedir: str, n_jobs: int = 8, skip_exist: bool = True):
        import os
        import glob
        from tqdm.auto import tqdm
        import joblib

        # create the directory if needed
        os.makedirs(basedir, exist_ok=True)

        if not skip_exist:
            orbits = glob.glob('*.EOF', root_dir=basedir)
            #print ('orbits', orbits)
            for orbit in orbits:
                os.remove(os.path.join(basedir, orbit))

        scenes = S1.scan_slc(basedir)
        mission_date = scenes[scenes.orbitpath.isnull()]\
            .reset_index()\
            .groupby(['date', 'mission'])\
            .first()\
            .reset_index()[['mission', 'datetime']]
        #print ('mission_date', mission_date)
        if len(mission_date) == 0:
            return

        access_token = self._get_copernicus_access_token()
        # download orbits
        with self.tqdm_joblib(tqdm(desc='ESA Downloading Sentinel-1 Orbits', total=len(mission_date))) as progress_bar:
            orbits = joblib.Parallel(n_jobs=n_jobs)(joblib.delayed(self._download_orbit)(access_token, date, mission, basedir)
                                           for (mission, date) in mission_date.values)
        return pd.Series(orbits)



# ----------------------------------------------------------------------------
# PyGMTSAR
# 
# This file is part of the PyGMTSAR project: https://github.com/mobigroup/gmtsar
# 
# Copyright (c) 2024, Alexey Pechnikov
# 
# Licensed under the BSD 3-Clause License (see LICENSE for details)
# ----------------------------------------------------------------------------
from .datagrid import datagrid
from .tqdm_joblib import tqdm_joblib

class GMT(datagrid, tqdm_joblib):
    
    # https://docs.generic-mapping-tools.org/6.0/datasets/earth_relief.html
    # only bicubic interpolation supported as the best one for the case
    def download_dem(self, geometry, filename=None, product='1s', skip_exist=True):
        """
        Download and preprocess SRTM digital elevation model (DEM) data using GMT library.

        Parameters
        ----------
        product : str, optional
            Product type of the DEM data. Available options are '1s' or 'SRTM1' (1 arcsec ~= 30m, default)
            and '3s' or 'SRTM3' (3 arcsec ~= 90m).

        Returns
        -------
        None or Xarray Dataarray

        Examples
        --------
        Download default STRM1 DEM (~30 meters):
        GMT().download_dem()

        Download STRM3 DEM (~90 meters):
        GMT.download_dem(product='STRM3')
        
        Download default STRM DEM to cover the selected area AOI:
        GMT().download_dem(AOI)
        
        Download default STRM DEM to cover all the scenes:
        GMT().download_dem(stack.get_extent().buffer(0.1))

        import pygmt
        # Set the GMT data server limit to N Mb to allow for remote downloads
        pygmt.config(GMT_DATA_SERVER_LIMIT=1e6)
        GMT().download_dem(AOI, product='1s')
        """
        import numpy as np
        import pygmt
        # suppress warnings
        pygmt.config(GMT_VERBOSE='errors')
        import os
        from tqdm.auto import tqdm

        if product in ['SRTM1', '1s', '01s']:
            resolution = '01s'
        elif product in ['SRTM3', '3s', '03s']:
            resolution = '03s'
        else:
            print (f'ERROR: unknown product {product}. Available only SRTM1 ("01s") and SRTM3 ("03s") DEM using GMT servers')
            return

        if filename is not None and os.path.exists(filename) and skip_exist:
            print ('NOTE: DEM file exists, ignore the command. Use "skip_exist=False" or omit the filename to allow new downloading')
            return

        lon_start, lat_start, lon_end, lat_end = self.get_bounds(geometry)
        with tqdm(desc='GMT SRTM DEM Downloading', total=1) as pbar:
            # download DEM using GMT extent W E S N
            dem = pygmt.datasets.load_earth_relief(resolution=resolution, region=[lon_start, lon_end, lat_start, lat_end])
            pbar.update(1)

        if filename is not None:
            if os.path.exists(filename):
                os.remove(filename)
            encoding = {'dem': self._compression(dem.shape)}
            dem.rename('dem').load().to_netcdf(filename, encoding=encoding, engine=self.netcdf_engine)
        else:
            return dem

    def download_landmask(self, geometry, filename=None, product='1s', skip_exist=True):
        """
        Download the landmask and save as NetCDF file.

        Parameters
        ----------
        product : str, optional
                Available options are '1s' (1 arcsec ~= 30m, default) and '3s' (3 arcsec ~= 90m).

        Examples
        --------
        from pygmtsar import GMT
        landmask = GMT().download_landmask(stack.get_dem())

        Notes
        -----
        This method downloads the landmask using GMT's local data or server.
        """
        import geopandas as gpd
        import numpy as np
        import pygmt
        # suppress warnings
        pygmt.config(GMT_VERBOSE='errors')
        import os
        #import subprocess
        from tqdm.auto import tqdm

        if filename is not None and os.path.exists(filename) and skip_exist:
            print ('NOTE: landmask file exists, ignore the command. Use "skip_exist=False" or omit the filename to allow new downloading')
            return

        if not product in ['1s', '3s']:
            print (f'ERROR: unknown product {product}. Available only "1s" or "3s" land mask using GMT servers')
            return

        lon_start, lat_start, lon_end, lat_end = self.get_bounds(geometry)
        with tqdm(desc='GMT Landmask Downloading', total=1) as pbar:
            landmask = pygmt.grdlandmask(resolution='f', region=[lon_start, lon_end, lat_start, lat_end], spacing=product, maskvalues='NaN/1')
            pbar.update(1)

        if filename is not None:
            if os.path.exists(filename):
                os.remove(filename)
            encoding = {'landmask': self._compression(landmask.shape)}
            landmask.rename('landmask').load().to_netcdf(filename, encoding=encoding, engine=self.netcdf_engine)
        else:
            return landmask
